/**   
 * @author Eugene Turkov
 * @author <a href="mailto:eugene.turkov@nasa.gov">eugene.turkov@nasa.gov</a>
 * @version: 1.0
 * @since SDK1.5
 * 
 * NASA Ames Research Center Computational Sciences Division Intelligent Data
 * Understanding Group (IDU) Moffett Field, CA 94035
 */

package turkov.io;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.Vector;

/**
 * This class is used to read / write CSV files.
 */
public class TUCSVFile extends File
{
	public TUCSVFile(final File file, final String valueSeparator)
			throws FileNotFoundException, IOException, Exception
	{
		this(new FileInputStream(file), valueSeparator, file.getAbsolutePath());
	}

	public TUCSVFile(final File file) throws FileNotFoundException,
			IOException, Exception
	{
		this(new FileInputStream(file), ",", file.getAbsolutePath());
	}

	public TUCSVFile(final InputStream inputStream,
			final String valueSeparator, final String absolutePath)
			throws IOException, Exception
	{
		super(absolutePath);
		this.valueSeparator = valueSeparator;
		this.absolutePath = absolutePath;

		BufferedReader bufferedReader = new BufferedReader(
				new InputStreamReader(inputStream));
		String line = "";
		dataLines = new Vector();
		while (line != null)
		{
			line = bufferedReader.readLine();
			if (line != null && !line.trim().equals(""))
			{
				String[] splitLine = line.split(valueSeparator);
				if (splitLine == null)
					dataLines.add(new String[] { line });
				else
					dataLines.add(splitLine);
			}
		}

		rows = dataLines.size();
		columns = 0;

		// do the below to determine the size of the data array
		if (((String[]) dataLines.get(0)).length > 0)
		{
			String[] fistRowColumns = (String[]) dataLines.get(0);

			if (fistRowColumns != null)
				columns = fistRowColumns.length;

			data = new String[rows][columns];

			loadData();
		}
	}

	private void loadData() throws Exception
	{
		try
		{
			for (int row = 0; row < rows; row++)
			{
				String[] columnDataForRow = (String[]) dataLines.get(row);
				for (int column = 0; column < columns; column++)
				{
					data[row][column] = columnDataForRow[column];
				}
			}
		}

		catch (ArrayIndexOutOfBoundsException e)
		{
			throw new Exception("The format of the csv file "
					+ this.getAbsolutePath() + " is invalid.");
		}
	}

	public String getString(int row, int column)
	{
		return data[row][column];
	}

	public Integer getInteger(int row, int column)
	{
		return new Integer(getString(row, column));
	}

	public Double getDouble(int row, int column)
	{
		return new Double(getString(row, column));
	}

	public Float getFloat(int row, int column)
	{
		return new Float(getString(row, column));
	}

	public int getRows()
	{
		return rows;
	}

	public int getColumns()
	{
		return columns;
	}

	public String toString()
	{
		StringBuffer resultBuffer = new StringBuffer();
		for (int row = 0; row < rows; row++)
		{
			for (int column = 0; column < columns; column++)
			{
				resultBuffer.append(this.getString(row, column));
				if (column < columns - 1)
					resultBuffer.append(COMMA);
			}
			if (row < rows - 1)
				resultBuffer.append(LINE_SEPARATOR);
		}
		return resultBuffer.toString();
	}

	private int rows;
	private int columns;
	private String[][] data;
	private Vector dataLines;
	private String absolutePath;
	private String valueSeparator;
	private final String COMMA = ",";
	private static final String LINE_SEPARATOR = System
			.getProperty("line.separator");
}