/** 
 * @author Eugene Turkov
 * @author <a href="mailto:eugene.turkov@nasa.gov">eugene.turkov@nasa.gov</a>
 * @version: 1.0
 * @since SDK1.5
 * 
 * NASA Ames Research Center Computational Sciences Division Intelligent Data
 * Understanding Group (IDU) Moffett Field, CA 94035
 */

package turkov.util;

import java.io.Serializable;
import java.util.List;
import java.util.Vector;

/**
 * This class is for representing simple sets NOT using the built-in java.util.TUSet
 * interface class. 
 */
public class TUSet implements Serializable
{
	/**
	 *     Creates a new instance of TUSet
	 *     
	 * 
	 * @param list the list which should be represented by this cluster.
	 */
	public TUSet(List list) throws UnsupportedOperationException
	{
		if (list == null)
			throw new UnsupportedOperationException("Can't create a cluster"
					+ " from a null list!");

		this.list = list;
	}

	/**
	 *     Creates a new instance of TUSet
	 */
	public TUSet()
	{
		this.list = new Vector();
	}

	/**
	 *     This method returns a String representation of the items represented
	 *     by this cluster.
	 *     
	 * 
	 * @param delimiter each element of this TUSet will be separated by the delimiter.
	 * @return a String representation of the items represented
	 *     by this cluster.
	 */
	public String toString(String delimiter)
	{
		StringBuffer resultStringBuffer = new StringBuffer();
		for (int i = 0; i < list.size(); i++)
		{
			resultStringBuffer.append(list.get(i).toString());
			if (i != list.size() - 1)
				resultStringBuffer.append(delimiter);
		}

		return resultStringBuffer.toString();
	}

	/**
	 This method returns a String representation of the items represented
	 by this cluster.
	 
	 @return a String representation of the items represented
	 by this cluster.
	 */
	public String toString()
	{
		return this.toString("\n");
	}

	/**
	 *     This method is used to add a new item to this TUSet.
	 *     
	 * 
	 * @param object the object being added to this TUSet.
	 */
	public void add(Object object)
	{
		if (!this.contains(object))
			list.add(object);
	}

	public Object remove(int i)
	{
		return list.remove(i);
	}

	/**
	 *     Returns true if this TUSet contains the object.
	 *     
	 * 
	 * @param object the object which may or may not be contained within this TUSet.
	 * @return false if the TUSet does not contain the specified object.
	 */
	public boolean contains(Object object)
	{
		return list.contains(object);
	}

	/**
	 *     This method is used to retrieve a particular item from this TUSet.
	 *     
	 * 
	 * @param i the item index to retrieve.
	 * @return the object located at index i.
	 */
	public Object get(int i)
	{
		return list.get(i);
	}

	public String[] toStringArray()
	{
		String[] result = new String[this.size()];
		for (int i = 0; i < result.length; i++)
			result[i] = this.get(i).toString();

		return result;
	}

	/**
	 *     Returns the number of elements in this TUSet.
	 *     
	 * 
	 * @return the number of elements in this TUSet.
	 */
	public int size()
	{
		return list.size();
	}

	/**
	 *     Use to determine if this TUSet is related to another TUSet (at least one element
	 *     is considered equal between the two sets).
	 *     
	 * 
	 * @param set the set with which this set should be compared
	 * @return true if this set has at least one item which appears in the
	 *     foreign set.
	 */
	public boolean hasAtLeastOneElementInCommonWith(TUSet set)
	{

		for (int i = 0; i < list.size(); i++)
		{
			Object localItem = list.get(i);
			if (set.contains(localItem))
				return true;
		}

		return false;
	}

	/**
	 Use this method to take the elements of one set and combine them with the
	 elements of another. After this operation is complete, this set will contain
	 the foreignSet as a subset.
	 
	 @param foreignSet the set with which this set's elements should be combined.
	 */
	public void mergeWith(TUSet foreignSet)
	{
		for (int i = 0; i < foreignSet.size(); i++)
		{
			this.add(foreignSet.get(i));
		}
	}

	/**
	 *     This method is used to add a new TUSet to this cluster. Sets contained
	 *     within this TUSet will be examined upon calling the add method.
	 *     If a cluster is found which is related to the foreignSet (cluster) then the
	 *     the add method will not add the foreignSet as a separate item in this TUSet,
	 *     it will however, merge the foreignSet with the TUSet that it is
	 *     related to in this collection. The result is the same number of Sets in this
	 *     TUSet, however with the merged cluster being larger now (possibly) and also several
	 *     clusters could become joined togther by the foreign cluster.
	 *     
	 * 
	 * @param foreignSet the item being added to this cluster (and if it is related
	 *     to any other items in this collection, the foreignSet and said item will be merged
	 *     into one.
	 */
	public void mergeContentsWith(TUSet foreignSet)
	{
		TUSet foreignCluster = foreignSet;
		TUSet initialCluster = null;
		boolean merged = false;
		for (int i = 0; i < this.size(); i++)
		{
			if (!(this.get(i) instanceof TUSet))
				throw new UnsupportedOperationException(
						"In order to merge contents,"
								+ " this set needs to be made-up of sets and the foreignSet must be"
								+ " a set.");

			TUSet localCluster = (TUSet) this.get(i);
			if (foreignCluster.hasAtLeastOneElementInCommonWith(localCluster))
			{
				if (initialCluster != null)
				{
					initialCluster.mergeWith(localCluster);
					list.remove(i);
					i--;
				}
				else
				{
					localCluster.mergeWith(foreignCluster);
					initialCluster = localCluster;
				}

				merged = true;
			}
		}

		if (!merged)
			add(foreignCluster);
	}

	/**
	 Use this method to compare sets. If two sets have the same elements (perhaps in
	 different order) then they are considered to be equal.
	 
	 @param object a foreign set object.
	 @return true if the foreign object is equal to this object.
	 */
	public boolean equals(Object object)
	{
		if (!(object instanceof TUSet))
			return false;
		TUSet set = (TUSet) object;
		if (set.size() != this.size())
			return false;

		for (int i = 0; i < set.size(); i++)
		{
			if (!this.contains(set.get(i)))
				return false;
		}

		return true;
	}

	protected List list;
}
