/**
Notices:

Copyright 2007 United States Government as represented by the Administrator of the National Aeronautics 
and Space Administration.  All Rights Reserved.

Disclaimers

No Warranty: THE SUBJECT SOFTWARE IS PROVIDED "AS IS" WITHOUT ANY WARRANTY OF ANY KIND, EITHER EXPRESSED, 
IMPLIED, OR STATUTORY, INCLUDING, BUT NOT LIMITED TO, ANY WARRANTY THAT THE SUBJECT SOFTWARE WILL CONFORM 
TO SPECIFICATIONS, ANY IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, OR FREEDOM 
FROM INFRINGEMENT, ANY WARRANTY THAT THE SUBJECT SOFTWARE WILL BE ERROR FREE, OR ANY WARRANTY THAT 
DOCUMENTATION, IF PROVIDED, WILL CONFORM TO THE SUBJECT SOFTWARE. THIS AGREEMENT DOES NOT, IN ANY MANNER, 
CONSTITUTE AN ENDORSEMENT BY GOVERNMENT AGENCY OR ANY PRIOR RECIPIENT OF ANY RESULTS, RESULTING DESIGNS, 
HARDWARE, SOFTWARE PRODUCTS OR ANY OTHER APPLICATIONS RESULTING FROM USE OF THE SUBJECT SOFTWARE.  FURTHER, 
GOVERNMENT AGENCY DISCLAIMS ALL WARRANTIES AND LIABILITIES REGARDING THIRD-PARTY SOFTWARE, IF PRESENT IN THE 
ORIGINAL SOFTWARE, AND DISTRIBUTES IT "AS IS."

Waiver and Indemnity:  RECIPIENT AGREES TO WAIVE ANY AND ALL CLAIMS AGAINST THE UNITED STATES GOVERNMENT, ITS 
CONTRACTORS AND SUBCONTRACTORS, AS WELL AS ANY PRIOR RECIPIENT.  IF RECIPIENT'S USE OF THE SUBJECT SOFTWARE 
RESULTS IN ANY LIABILITIES, DEMANDS, DAMAGES, EXPENSES OR LOSSES ARISING FROM SUCH USE, INCLUDING ANY DAMAGES 
FROM PRODUCTS BASED ON, OR RESULTING FROM, RECIPIENT'S USE OF THE SUBJECT SOFTWARE, RECIPIENT SHALL INDEMNIFY 
AND HOLD HARMLESS THE UNITED STATES GOVERNMENT, ITS CONTRACTORS AND SUBCONTRACTORS, AS WELL AS ANY PRIOR RECIPIENT, 
TO THE EXTENT PERMITTED BY LAW.  RECIPIENT'S SOLE REMEDY FOR ANY SUCH MATTER SHALL BE THE IMMEDIATE, UNILATERAL 
TERMINATION OF THIS AGREEMENT.
-------------------------------------------------------------------------------------------------------------------
*/


/**
 SequenceMiner
 
 @author: Eugene Turkov, {eturkov@email.arc.nasa.gov}
 @version: 1.0
 
 Finds anomalous sequences and characterizes why they are anomalous.
 
 NASA Ames Research Center
 Computational Sciences Division
 Intelligent Data Understanding Group (IDU)
 Moffett Field, CA 94035
 */
package sequenceMiner;
import java.io.IOException;
import java.io.File;
import turkov.lang.TUStringBuilder;
import turkov.util.TUClusterUtil;
import turkov.math.algorithms.TUClusterableItem;
import turkov.math.algorithms.TUCluster;
import turkov.math.algorithms.TUDistanceMeasures;
import turkov.math.algorithms.TUClusteringMethods;
import turkov.util.TUSequenceAlteration;
import java.util.Arrays;
import turkov.io.TUTextFileWriter;
import java.util.Vector;
import java.util.HashMap;
import turkov.io.TUCSVFile;
import turkov.util.TUSet;

/**
 This is the main program class file
 */
public class SequenceMiner
{
    /**
     This method is called when it is neccessary to printToStdError application useage.
     
     
     @param errorMessage the error message that will be displayed if this method
     deems neccessary.
     */
    private static void printUseage(String errorMessage)
    {
        println();
        println(errorMessage);
        System.out.println(
"Sequence Miner Version 1.1 USER COMMANDS:" +
        "\n\n" +
        "\t-h" +
                "\n\t\tShow this help message." +
        "\n\n" +
	"\t-o clusteringOutputFile" +
	"\n\n" +
		"\t\tIf this option is not specified then a deafult output" +
		"\n\t\tfile will be created in the current working directory by"  +
		"\n\t\tthe name of clusters.csv" +
        "\n\n" +
	"\t-i inputFile" +
	"\n\n" +
		"\t\tThe input file must be specified (either absolute or" +
		"\n\t\trelative path is okay). The input file must contain" +
		"\n\t\tsequences of characters. Each line in the file represents" +
		"\n\t\ta sequence where each space denotes a separation between" +
                "\n\t\tsequence elements." +
        "\n\n" +
	"\t-k" +

		"\n\t\tSpecify the number of CLARA clusters to use. The deafult" +
                "\n\t\tvalue is 4" +
        "\n\n" +
	"\t-n" +

		"\n\t\tSpeicfy the number of PAM iterations." +
                "\n\t\tThe default value is 5" +
        "\n\n" +
	"\t-c" +
	
		"\n\t\tSpecify the number of CLARA iterations. The default" +
                "\n\t\tvalue is 5" +
        "\n\n" +
	"\t-s" +

		"\n\t\tSpecify the sample fraction. The sample fraction is the" +
                "\n\t\tfraction of total data to be used when running PAM. The" +
                "\n\t\tdefault value is 0.1" +
        "\n\n" +
	"\t-f" +

		"\n\t\tSpeicfy the outlier fraction. The outlier fraction" +
                "\n\t\tspecifies the tcutoff for what is considered an outlier." +
                "\n\t\tThe default value is 0.1" +
        "\n\n" +
	"\t-x" +

		"\n\t\tSpecify a file name prefix for files containing suggested" +
                "\n\t\tsequence alterations." +
        "\n\n" +
	"\t-a" +

		"\n\t\tIf this option is specified, only clutering will be" +
                "\n\t\tperformed." +
        "\n\n" +
	"\t-p predeterminedClustersCSVFile" +
                "\n" +
		"\n\t\tSpecify this option to skip clustering and use clustering" +
                "\n\t\tinformation from file. The format of this file should be" +                
                "\n\t\tthe same as SequenceMiner's clustering results file." +
        "\n\n" +
	"\t-e sequenceIdFile" +
                "\n" +
		"\n\t\tSpecify this option to ignore the top outliers and use" +
                "\n\t\tsequences with these IDs instead." +   
        "\n\n" +
	"\t-u supplementalSequencesFile" +
                "\n" +
		"\n\t\tSpecify this option to ignore the top outliers and use" +
                "\n\t\tthese new sequences instead." +                 
"\n\n" +
"NOTES:" +
	"\n\tSequence alteration csv file format:" +
		"\n\t\tcolumn 1: 1-based index" +
		"\n\t\tcolumn 2: element" +
		"\n\t\tcolumn 3: score" +
        "\n\n" +    
	"\tClustering csv file format:" +
		"\n\t\tcolumn 1: sequence id (the id is 1-based and determined" +
                "\n\t\tby the order in which the sequence was read)." +
		"\n\n" +
		"\t\tcolumn 2: owner cluster id (the owner cluster id is 1-based" +
                        "\n\t\t\tand states to which cluster this sequence belongs)." +
		"\n\n" +
		"\t\tcolumn 3: outlier score" +
		"\n\n" +
		"\t\tcolumn 4: rank (rank is determined by outlier score. Items" +
                        "\n\t\t\twith the highest outlier scores have the highest" +
                        "\n\t\t\trank. For example, the item with the highest" +
                        "\n\t\t\toutlier score would have rank 1)." +
		"\n\n" +
		"\t\tcolumn 5: this column contains a 1 if the current sequence" +
                        "\n\t\t\thas a high enough rank to be looked at as an" +
                        "\n\t\t\toutlier, and 0 otherwise.");                            
    }
    
    /**
     This is the main method for calling the application.
     
     @param args a string array containing input arguments.
     @throws IOException if there is a problem reading or writing data.
     @throws Exception if there are any other unforseen problems.
     */
    public static void main(String[] args)
    throws IOException
    {
        try
        {
            long startTime = System.currentTimeMillis();
            // an input file to read sequences from
            inputFile = null;
            
            // an output file to write the output to
            clusteringOutputFile = null;
            
            /* process all of the input arguments to initialize the above
             variables as neccessary */
            processInputArguments(args);
            
            initializeIOFileSetup();
            
            ensureSafeToProceed();
            
            long a = 0;
            
            if(inputFile != null)
            {
                a = System.currentTimeMillis();
                loadSequenceData();
            System.err.println("Processed sequence data from input file in: "
                    + TUStringBuilder.millisecondsToHoursMinutesSeconds
                    (System.currentTimeMillis() - a));                
                
            }
                                                  
            a = System.currentTimeMillis();
            
            if(predeterminedClustersCSVFile == null)
                doClustering();
            
            else
            {
                clusters = TUClusteringMethods.getClaraClusters(predeterminedClustersCSVFile, inputFile);
            }                        
            
            System.err.println("Clustering segment completed in: "
                    + TUStringBuilder.millisecondsToHoursMinutesSeconds
                    (System.currentTimeMillis() - a));
            
            
            if(supplementalSequencesFile != null)
            {
                a = System.currentTimeMillis();
                loadSupplementalSequenceData();
                System.err.println("Processed supplemental sequence data in: "
                        + TUStringBuilder.millisecondsToHoursMinutesSeconds
                        (System.currentTimeMillis() - a));
            }
            
            a = System.currentTimeMillis();
            initializeMappings();
            System.err.println("Sequence mappings initialized in: "
                    + TUStringBuilder.millisecondsToHoursMinutesSeconds
                    (System.currentTimeMillis() - a));            
            
            if(!onlyDoClustering)
            {
                a = System.currentTimeMillis();
                writeMissingAndExtraElements();
                System.err.println("Sequence alteration segment completed in: "
                        + TUStringBuilder.millisecondsToHoursMinutesSeconds
                        (System.currentTimeMillis() - a));
            }
            
            else
                System.err.println("Skipping sequence alteration segment.");
            
            if(supplementalSequencesFile != null)
            {
                a = System.currentTimeMillis();
                TUClusterUtil.setRankValues(clusters);
                long b = System.currentTimeMillis();
                println("Rank values set in: " + TUStringBuilder.millisecondsToHoursMinutesSeconds(b - a));

                a = System.currentTimeMillis();
                TUClusterUtil.writeOutputCSVFile(clusters, clusteringOutputFile, true
                        , outlierFraction, true);
                b = System.currentTimeMillis();
                println("Wrote " + clusteringOutputFile.getAbsolutePath() + " in "
                        + TUStringBuilder.millisecondsToHoursMinutesSeconds(b - a));
            }
            
            System.err.println("SequenceMiner finished in: "
                    + TUStringBuilder.millisecondsToHoursMinutesSeconds
                    (System.currentTimeMillis() - startTime));
        }

        catch(IOException e)
        {
            println(e.getMessage());
            //e.printStackTrace();
        }
        
    }
    
    /**
     This method is responsible for creating sequence alteration files for each
     sequence.
     
     @throws IOException if there is a problem writing to a file
     */
    private static void writeMissingAndExtraElements()
    throws IOException
    {
        if(clusters == null)
            return;
        
        int[] sequenceIds = null;
        
        if(sequenceIdFile != null)
            sequenceIds = getSequenceIdList(sequenceIdFile);
        
        for(int i = 0; i < clusters.length; i++)
        {
            TUCluster cluster = clusters[i];
            for(int j = 0; j < cluster.size(); j++)
            {
                TUClusterableItem item = (TUClusterableItem)cluster.get(j);               
                
                
                if(sequenceIds != null)
                {
                    // then use the sequences specified, don't do anything else
                    for(int k = 0; k < sequenceIds.length; k++)
                    {
                        if(sequenceIds[k] == item.getId())
                        {                 
                            //System.out.println(item.getId());
                            Object[] outlierSequence = (Object[])item.getValue();                                 

                            TUSequenceAlteration[] suggestedAdditions
                                    = TUClusterUtil.getMissingElementsInOrder(outlierSequence,
                                    cluster, uniqueElementsToIds, idsToUniqueElements, item.getId());                                        

                            TUSequenceAlteration[] suggestedRemovals
                                    = TUClusterUtil.getNonEssentialElements(outlierSequence,
                                    cluster, uniqueElementsToIds, idsToUniqueElements, item.getId());

                            Arrays.sort(suggestedRemovals);

                            writeAlterationsFile(item, suggestedAdditions, "missingElements.csv");
                            writeAlterationsFile(item, suggestedRemovals, "extraElements.csv"); 
                            break;
                        }
                    }
                }
                
                else if(supplementalSequencesFile != null)
                {                    
                    // only print out the new sequences that were added in this file
                    if(item.getCustomData(turkov.util.TUClusterUtil.SUPPLEMENTAL_ITEM) != null)
                    {
                        Object[] outlierSequence = (Object[])item.getValue();                                 

                        TUSequenceAlteration[] suggestedAdditions
                                = TUClusterUtil.getMissingElementsInOrder(outlierSequence,
                                cluster, uniqueElementsToIds, idsToUniqueElements, item.getId());                                        

                        TUSequenceAlteration[] suggestedRemovals
                                = TUClusterUtil.getNonEssentialElements(outlierSequence,
                                cluster, uniqueElementsToIds, idsToUniqueElements, item.getId());

                        Arrays.sort(suggestedRemovals);

                        writeAlterationsFile(item, suggestedAdditions, "missingElements.csv");
                        writeAlterationsFile(item, suggestedRemovals, "extraElements.csv");                        
                    }
                }
                
                else if(predeterminedClustersCSVFile != null && !userSpecifiedAnOutlierFraction)
                {
                    // use predetermind ranking info to print outliers
                    if(item.getCustomData(turkov.util.TUClusterUtil.OUTLIER) != null)
                    {
                        Object[] outlierSequence = (Object[])item.getValue();                                 

                        TUSequenceAlteration[] suggestedAdditions
                                = TUClusterUtil.getMissingElementsInOrder(outlierSequence,
                                cluster, uniqueElementsToIds, idsToUniqueElements, item.getId());                                        

                        TUSequenceAlteration[] suggestedRemovals
                                = TUClusterUtil.getNonEssentialElements(outlierSequence,
                                cluster, uniqueElementsToIds, idsToUniqueElements, item.getId());

                        Arrays.sort(suggestedRemovals);

                        writeAlterationsFile(item, suggestedAdditions, "missingElements.csv");
                        writeAlterationsFile(item, suggestedRemovals, "extraElements.csv");                        
                    }                    
                }
                
                else if(item.getRank() <=  ((double)sequences.length * outlierFraction))
                {
                    // actually pick items out according to rank (using outlierFraction info)
                    Object[] outlierSequence = (Object[])item.getValue();                                 
                    
                    TUSequenceAlteration[] suggestedAdditions
                            = TUClusterUtil.getMissingElementsInOrder(outlierSequence,
                            cluster, uniqueElementsToIds, idsToUniqueElements, item.getId());                                        
                    
                    TUSequenceAlteration[] suggestedRemovals
                            = TUClusterUtil.getNonEssentialElements(outlierSequence,
                            cluster, uniqueElementsToIds, idsToUniqueElements, item.getId());
                                       
                    Arrays.sort(suggestedRemovals);
                    
                    writeAlterationsFile(item, suggestedAdditions, "missingElements.csv");
                    writeAlterationsFile(item, suggestedRemovals, "extraElements.csv");
                }
            }
        }
    }
    
    /**
     This method is responsible for invoking any clusering in SequenceMiner.
     @throws IOException if there are any problems writing to the output file.
     */
    private static void doClustering()
    throws IOException
    {
        long a = System.currentTimeMillis();
        clusters = TUClusteringMethods.getClaraClusters(numberOfClaraClusters,
                clusterableItems, TUDistanceMeasures.LONGEST_COMMON_SUBSEQUENCE,
                numberOfPAMIterations, numberOfCLARAIterations, sampleFraction);
        
        //System.out.println("clusters: " + clusters);
        
        // get top outliers ( for now assume everything is an outlier - sequences)
        //given the cluster it belongs to and the sequence , call the extra valuse fucntion
        
        System.err.println("Final iteration CLARA Clustering Quality: "
                + TUClusteringMethods.getClusteringQuality(clusters
                , TUDistanceMeasures.LONGEST_COMMON_SUBSEQUENCE));
        
        long b = System.currentTimeMillis();
        println("CLARA clustering ran for: " + TUStringBuilder.millisecondsToHoursMinutesSeconds(b - a));
        
        if(supplementalSequencesFile == null)
        {
            a = System.currentTimeMillis();
            TUClusterUtil.setRankValues(clusters);
            b = System.currentTimeMillis();
            println("Rank values set in: " + TUStringBuilder.millisecondsToHoursMinutesSeconds(b - a));

            a = System.currentTimeMillis();
            TUClusterUtil.writeOutputCSVFile(clusters, clusteringOutputFile, true,
                    outlierFraction, false);
            b = System.currentTimeMillis();
            println("Wrote " + clusteringOutputFile.getAbsolutePath() + " in "
                    + TUStringBuilder.millisecondsToHoursMinutesSeconds(b - a));
        }
        
    }
    
    /**
     This method is used to create a file for each sequence where the contents
     of the file will indicate some alteration that could be done to the sequence.
     
     @param item the TUClusterableItem whose sequence (value) is to be used to
     suggest alterations for.
     @param alterations the suggested list of alterations for the sequence belonging
     to item.
     @param fileNameSuffix the suffix of the file name of the file to be created.
     */
    private static void writeAlterationsFile(TUClusterableItem item
            , TUSequenceAlteration[] alterations, String fileNameSuffix)
            throws IOException
    {
        final int offset = 1;

        //System.err.println("Writing out: " + fileNameSuffix + " which has " + alterations.length + " elements.");
        StringBuffer stringBuffer = new StringBuffer();
        for(int i = 0; i < alterations.length; i++)
        {
            stringBuffer.append((alterations[i].getIndex() + offset) + ","
                    + alterations[i].getObject()
                    + "," + alterations[i].getScore());
             
            stringBuffer.append(System.getProperty("line.separator"));
        }
        
        TUTextFileWriter.createFile(new File(alterationsFileNamePrefix + "_"
                + (item.getId() + offset)
                + "_" + fileNameSuffix)
                , new String[]{stringBuffer.toString()});
    }
    
    /**
     This method's sole responsibility is to configure File input and output.
     
     @param inputFile the file from which data shall be read.
     @param clusteringOutputFile the file to which data shall be written.
     @throws IOException if there is a problem reading or writing a file.
     */
    private static void initializeIOFileSetup()
    throws IOException
    {
        if(clusteringOutputFile != null && clusteringOutputFile.exists())
            clusteringOutputFile.delete();
        
        if(inputFile != null)
        {
            long a = System.currentTimeMillis();
            sequences = TUStringBuilder.generateCollectionFromFile(inputFile);
            System.err.println("Read-in sequence data from sequencesFile in: "
                    + TUStringBuilder.millisecondsToHoursMinutesSeconds
                    (System.currentTimeMillis() - a));              
        }
        
        if(supplementalSequencesFile != null)
        {
            long a = System.currentTimeMillis();
            supplementalSequences = TUStringBuilder.generateCollectionFromFile(supplementalSequencesFile);
            System.err.println("Read-in supplemental sequence data from supplementalSequencesFile in: "
                    + TUStringBuilder.millisecondsToHoursMinutesSeconds
                    (System.currentTimeMillis() - a));             
        }
            
            
    }
    
    /**
     This method's sole purpose is to take the input arguments and break them up
     amongs their appropriate variables as neccessary.
     
     @param args the input arguments
     */
    private static void processInputArguments(String[] args)
    throws IOException
    {
        // unused options: b d g j l m q r t v w y z 0 1 2 3 4 5 6 7 8 9
        
        if(args.length == 0)
        {
            printUseage("ERROR: No input arguments detected!");
            System.exit(1);
        }
        
        for(int i = 0; i < args.length; i++)
        {
            if(args[i].startsWith("-"))
            {
                if(args[i].length() != 2)
                {
                    if(args[i].length() == 1)
                    {
                        printUseage("You failed to specify an argument immediately" +
                                " after a dash(-).");
                    }
                                            
                    else
                    {
                        printUseage("Unrecognized argument enetered: " + args[i]);
                    }
                    System.exit(1);
                }
                
                if(args[i].toLowerCase().charAt(1) == 'i')
                {
                    if(args.length - 1 < i + 1)
                    {
                        printUseage(args[i] + " not followed by input argument.");
                        System.exit(1);
                    }
                    
                    else
                    {
                        inputFile = new File(args[i + 1]);
                        if(!inputFile.exists())
                        {
                            println("The specified file " +
                                    inputFile.getAbsolutePath() + " does not exist.");
                            System.exit(1);
                        }
                        
                        else if(inputFile.isDirectory())
                        {
                            println("You specified a directory for input. This" +
                                    " feature is not yet supported.");
                            System.exit(1);
                        }
                        
                        else
                        {
                            println("Using input file: " + inputFile.getAbsolutePath());
                        }
                        
                        i++;
                    }
                }
                
                else if(args[i].toLowerCase().charAt(1) == 'p')
                {
                    if(args.length - 1 < i + 1)
                    {
                        printUseage(args[i] + " not followed by input argument.");
                        System.exit(1);
                    }
                    
                    else
                    {
                        try
                        {
                            predeterminedClustersCSVFile = new TUCSVFile(new File(args[i + 1]));
                        }
                        
                        catch(Throwable t)
                        {
                            println(t.getMessage());
                        }
                        
                        if(!predeterminedClustersCSVFile.exists())
                        {
                            println("The specified file " +
                                    predeterminedClustersCSVFile.getAbsolutePath()
                                    + " does not exist.");
                            
                            System.exit(1);
                        }
                        
                        else if(predeterminedClustersCSVFile.isDirectory())
                        {
                            println("You specified a directory as a predetermined" +
                                    "ClustersCSVFile. This" +
                                    " feature is not yet supported.");
                            
                            System.exit(1);
                        }
                        
                        else
                        {
                                println("Using predeterminedClustersCSVFile file: "
                                    + predeterminedClustersCSVFile.getAbsolutePath());
                        }
                        
                        i++;
                    }
                }                
                
                else if(args[i].toLowerCase().charAt(1) == 'e')
                {
                    if(args.length - 1 < i + 1)
                    {
                        printUseage(args[i] + " not followed by input argument.");
                        System.exit(1);
                    }
                    
                    else
                    {
                        
                        sequenceIdFile = new File(args[i + 1]);
                        
                        if(!sequenceIdFile.exists())
                        {
                            println("The specified file " +
                                    sequenceIdFile.getAbsolutePath()
                                    + " does not exist.");
                            
                            System.exit(1);
                        }
                        
                        else if(sequenceIdFile.isDirectory())
                        {
                            println("You specified a directory as a sequenceIdFile. This" +
                                    " feature is not yet supported.");
                            
                            System.exit(1);
                        }
                        
                        else if(TUStringBuilder.generateFromFile(sequenceIdFile).trim().equals(""))
                        {
                            println("You specified and empty sequenceIdFile.");
                            
                            System.exit(1);                            
                        }
                        
                        else
                        {
                            println("Using sequenceIdFile file: "
                                    + sequenceIdFile.getAbsolutePath());
                        }
                        
                        i++;
                    }
                }     
                
                else if(args[i].toLowerCase().charAt(1) == 'u')
                {
                    if(args.length - 1 < i + 1)
                    {
                        printUseage(args[i] + " not followed by input argument.");
                        System.exit(1);
                    }
                    
                    else
                    {                        
                        supplementalSequencesFile = new File(args[i + 1]);
                        
                        if(!supplementalSequencesFile.exists())
                        {
                            println("The specified file " +
                                    supplementalSequencesFile.getAbsolutePath()
                                    + " does not exist.;l");
                            
                            System.exit(1);
                        }
                        
                        else if(supplementalSequencesFile.isDirectory())
                        {
                            println("You specified a directory as a supplementalSequencesFile. This" +
                                    " feature is not yet supported.");
                            
                            System.exit(1);
                        }
                        
                        else
                        {
                            println("Using supplementalSequencesFile file: "
                                    + supplementalSequencesFile.getAbsolutePath());
                        }
                        
                        i++;
                    }
                }                           
                
                else if(args[i].toLowerCase().charAt(1) == 'h')
                {
                    printUseage("");
                    System.exit(0);
                }
                
                else if(args[i].toLowerCase().charAt(1) == 'o')
                {
                    if(args.length - 1 < i + 1)
                    {
                        printUseage(args[i] + " not followed by input argument.");
                        System.exit(1);
                    }
                    
                    else
                    {
                        clusteringOutputFile = new File(args[i + 1]);
                        println("Using output file: " + clusteringOutputFile.getAbsolutePath());
                    }
                    i++;
                }
                
                else if( args[i].length() == 2 && args[i].substring(1).toLowerCase().equals("n"))
                {
                    if(args.length - 1 < i + 1)
                    {
                        printUseage(args[i] + " not followed by input argument.");
                        System.exit(1);
                    }
                    
                    if(numberOfPAMIterations > 0)
                    {
                        printUseage("The number of PAM iterations has already been specified.");
                        System.exit(1);
                    }
                    
                    else
                    {
                        try
                        {
                            numberOfPAMIterations = Integer.parseInt(args[i + 1]);
                        }
                        
                        catch(NumberFormatException e)
                        {
                            println("ERROR: Please specify a valid positive" +
                                    " integer (greater than zero)" +
                                    " for the numberOfPAMIterations.");
                            System.exit(1);
                        }
                        
                        if(numberOfPAMIterations < 1)
                        {
                            println("ERROR: Please specify a valid positive" +
                                    " integer (greater than zero)" +
                                    " for the numberOfPAMIterations.");
                            System.exit(1);
                        }
                        
                        i++;
                    }
                }
                
                else if( args[i].length() == 2 && args[i].substring(1).toLowerCase().equals("c"))
                {
                    if(args.length - 1 < i + 1)
                    {
                        printUseage(args[i] + " not followed by input argument.");
                        System.exit(1);
                    }
                    
                    if(numberOfCLARAIterations > 0)
                    {
                        printUseage("The number of CLARA iterations has already been specified.");
                        System.exit(1);
                    }
                    
                    else
                    {
                        try
                        {
                            numberOfCLARAIterations = Integer.parseInt(args[i + 1]);
                        }
                        
                        catch(NumberFormatException e)
                        {
                            println("ERROR: Please specify a valid positive" +
                                    " integer (greater than zero)" +
                                    " for the numberOfCLARAIterations.");
                            System.exit(1);
                        }
                        
                        if(numberOfCLARAIterations < 1)
                        {
                            println("ERROR: Please specify a valid positive" +
                                    " integer (greater than zero)" +
                                    " for the numberOfCLARAIterations.");
                            System.exit(1);
                        }
                        
                        i++;
                    }
                }
                
                else if( args[i].length() == 2 && args[i].substring(1).toLowerCase().equals("k"))
                {
                    if(args.length - 1 < i + 1)
                    {
                        printUseage(args[i] + " not followed by input argument.");
                        System.exit(1);
                    }
                    
                    if(numberOfClaraClusters > 0)
                    {
                        printUseage("The number of CLARA clusters has already been specified.");
                        System.exit(1);
                    }
                    
                    else
                    {
                        try
                        {
                            numberOfClaraClusters = Integer.parseInt(args[i + 1]);
                        }
                        
                        catch(NumberFormatException e)
                        {
                            println("ERROR: Please specify a valid positive non-zero" +
                                    " integer" +
                                    " for the numberOfClaraClusters.");
                            System.exit(1);
                        }
                        
                        if(numberOfClaraClusters < 1)
                        {
                            println("ERROR: Please specify a valid positive non-zero" +
                                    " integer" +
                                    " for the numberOfClaraClusters.");
                            System.exit(1);
                        }
                        
                        i++;
                    }
                }
                
                else if( args[i].length() == 2 && args[i].substring(1).toLowerCase().equals("s"))
                {
                    if(args.length - 1 < i + 1)
                    {
                        printUseage(args[i] + " not followed by input argument.");
                        System.exit(1);
                    }
                    
                    if(sampleFraction > 0)
                    {
                        printUseage("The sample percentage has already been specified.");
                        System.exit(1);
                    }
                    
                    else
                    {
                        try
                        {
                            sampleFraction = Float.parseFloat(args[i + 1]);
                        }
                        
                        catch(NumberFormatException e)
                        {
                            println("ERROR: Please specify a positive" +
                                    " value x, such that 0 < x <= 1 (" +
                                    " for the sampleFraction.");
                            System.exit(1);
                        }
                        
                        if(sampleFraction <= 0 || sampleFraction > 1)
                        {
                            println("ERROR: Please specify a positive" +
                                    " value x, such that 0 < x <= 1 (" +
                                    " for the sampleFraction.");
                            System.exit(1);
                        }
                        i++;
                    }
                }
                
                else if( args[i].length() == 2 && args[i].substring(1).toLowerCase().equals("f"))
                {
                    if(args.length - 1 < i + 1)
                    {
                        printUseage(args[i] + " not followed by input argument.");
                        System.exit(1);
                    }
                    
                    if(outlierFraction > 0)
                    {
                        printUseage("The outlier fraction has already been specified.");
                        System.exit(1);
                    }
                    
                    else
                    {
                        try
                        {
                            outlierFraction = Float.parseFloat(args[i + 1]);
                            userSpecifiedAnOutlierFraction = true;
                        }
                        
                        catch(NumberFormatException e)
                        {
                            println("ERROR: Please specify a positive" +
                                    " value x, such that 0 < x <= 1 (" +
                                    " for the outlierFraction.");
                            System.exit(1);
                        }
                        
                        if(outlierFraction <= 0 || outlierFraction > 1)
                        {
                            println("ERROR: Please specify a positive" +
                                    " value x, such that 0 < x <= 1 (" +
                                    " for the outlierFraction.");
                            System.exit(1);
                        }
                        i++;
                    }
                }
                
                else if(args[i].toLowerCase().charAt(1) == 'x')
                {
                    if(args.length - 1 < i + 1)
                    {
                        printUseage(args[i] + " not followed by input argument.");
                        System.exit(1);
                    }
                    
                    else
                    {
                        alterationsFileNamePrefix = args[i + 1];
                    }
                    i++;
                }
                
                
                else if( args[i].length() == 2 && args[i].substring(1).toLowerCase().equals("a"))
                {
                    if(onlyDoClustering)
                    {
                        printUseage("onlyDoClustering mode has already been specified.");
                        System.exit(1);
                    }
                    
                    onlyDoClustering = true;
                }
                
                else
                {
                    printUseage("Unrecognized argument entered: " + args[i]);
                    System.exit(1);
                }
            }           
        }
    }
    
    /**
     This method will print the given string to standard error and then print a
     new line immediately after.
     
     @param s the string to be printed.
     */
    private static void println(String s)
    {
        printToStdError(s);
        println();
    }
    
    /**
     This method will print a new line to standard error.
     */
    private static void println()
    {
        printToStdError(System.getProperty("line.separator"));
    }
    
    private static void printToStdError(String s)
    {
        System.err.print(s);
        System.err.flush();
    }
    
    /**
     This method confirms that all preliminary settings have been activated and
     that every thing is GREEN to go. If certain settings haven't been specified
     then defaults (if available) are used, otherwise and exception will be thrown.
     */
    private static void ensureSafeToProceed()
    {
        if(alterationsFileNamePrefix == null)
            alterationsFileNamePrefix = DEFAULT_ALTERATIONS_FILE_NAME_PREFIX;
        
        if(numberOfPAMIterations <= 0)
            numberOfPAMIterations = DEFAULT_NUMBER_OF_PAM_ITERATIONS;
        
        if(numberOfClaraClusters <= 0)
            numberOfClaraClusters = DEFAULT_NUMBER_OF_CLUSTERS;
        
        if(numberOfCLARAIterations <= 0)
            numberOfCLARAIterations = DEFAULT_NUMBER_OF_CLARA_ITERATIONS;
        
        if(sampleFraction <= 0)
            sampleFraction = DEFAULT_SAMPLE_FRACTION;
        
        if(outlierFraction <= 0)
            outlierFraction = DEFAULT_OUTLIER_FRACTION;
        
        if(clusteringOutputFile == null)
            clusteringOutputFile = new File("clusters.csv");
        
        if(inputFile == null && supplementalSequencesFile == null)
        {
            printUseage("Neither the inputFile (sequences) nor the supplemental" +
                    "SequencesFile have been specified; unable to proceed.");
            System.exit(1);
        }
        
        if(supplementalSequencesFile != null && sequenceIdFile != null)
        {
            printUseage("You specified both a supplementalSequencesFile " +
                    "and a sequenceIdFile; this is not allowed - please choose one.");
            System.exit(1);
        }
        
        if(predeterminedClustersCSVFile != null && onlyDoClustering)        
        {
            printUseage("It doesn't make" +
                    " sense to request just clustering when " +
                    "you are providing your own clustering " +
                    "results!");
            System.exit(1);
        }                                   

    }
    
    /**
     This method is for initializing the mappings to be used by various methods
     within sequence miner. These mappings map from uniqueElements to their IDs
     and the other way around too. A unique element can be considered as follows:
     
     If a sequence consists of elements and each element is a letter of the
     alphabet, then the following mapping is reasonable:
     
     a --> 0
     b --> 1
     c --> 2
     .
     .
     .
     etc...
     
     */
    private static void initializeMappings()
    {
        uniqueElementsToIds = new HashMap();
        idsToUniqueElements = new HashMap();
        
        int count = 0;
        for(int i = 0; i < clusterableItems.length; i++)
        {
            Object[] sequence = (Object[])clusterableItems[i].getValue();
            
            /*
             System.out.println("the current sequence is: "
             + TUStringBuilder.concatenate(sequence," "));
             */
            
            for(int j = 0; j < sequence.length; j++)
            {
                if(uniqueElementsToIds.get(sequence[j]) != null)
                    continue;
                
                Integer integer = new Integer(count);
                
                //System.out.println("item: " + sequence[j] + " will have ID: " + integer);
                uniqueElementsToIds.put(sequence[j], integer);
                idsToUniqueElements.put(integer, sequence[j]);
                count++;
            }
        }
        
        if(supplementalClusterableItems != null)
        {
            for(int i = 0; i < supplementalClusterableItems.length; i++)
            {
                Object[] sequence = (Object[])supplementalClusterableItems[i].getValue();

                /*
                 System.out.println("the current sequence is: "
                 + TUStringBuilder.concatenate(sequence," "));
                 */

                for(int j = 0; j < sequence.length; j++)
                {
                    if(uniqueElementsToIds.get(sequence[j]) != null)
                        continue;

                    Integer integer = new Integer(count);

                    //System.out.println("item: " + sequence[j] + " will have ID: " + integer);
                    uniqueElementsToIds.put(sequence[j], integer);
                    idsToUniqueElements.put(integer, sequence[j]);
                    count++;
                }
            }        
        }
    }
    
    /**
     This method is called to load all of the sequences from source.
     */
    private static void loadSequenceData()
    {
        /* using a vector initially because we don't know how many sequences we
        will be working with */
        
        Vector temp = new Vector();
        
        for(int i = 0; i < sequences.length; i++)
        {           
            sequences[i] = sequences[i].trim();
            String[] sequenceData = sequences[i].split(" ");
            
            /*
            System.out.println("just read in: "
            + TUStringBuilder.concatenate(sequenceData, "<space>"));
             */
            
            if(sequenceData.length == 0
                    || (sequenceData.length == 1
                    && sequenceData[0].equals("")))
                continue;
            
            Vector tempSequence = new Vector();
            for(int j = 0; j < sequenceData.length; j++)
            {
                try
                {
                    if(sequenceData[j].trim().equals(""))
                        continue;
                    
                    //System.out.println("adding: " + sequenceData[j]);
                    tempSequence.add(String.valueOf(sequenceData[j]));
                    //sequence[j] = Integer.valueOf(sequenceData[j]);
                }
                catch(NumberFormatException e)
                {
                    System.err.println("Read in a bad element in the input file" +
                            " at line: " + i + " check the: " + (j + 1) + "th item.");
                }                
            }
            
            /* convert the vector back over to an array */
            String[] sequence = new String[tempSequence.size()];
            for(int k = 0; k < tempSequence.size(); k++)
                sequence[k] = (String)tempSequence.get(k);
            
            temp.add(new TUClusterableItem(sequence, i));
        }
        
        /* convert the vector back over to an array */
        clusterableItems = new TUClusterableItem[temp.size()];
        for(int i = 0; i < temp.size(); i++)
            clusterableItems[i] = (TUClusterableItem)temp.get(i);
    }
    
    /**
     This method is called to load all of the sequences from source.
     */
    private static void loadSupplementalSequenceData()
    {
        /* using a vector initially because we don't know how many sequences we
        will be working with */
        
        Vector temp = new Vector();
        
        for(int i = 0; i < supplementalSequences.length; i++)
        {
            long a = System.currentTimeMillis();
            supplementalSequences[i] = supplementalSequences[i].trim();
            String[] sequenceData = supplementalSequences[i].split(" ");
            
            /*
            System.out.println("just read in: "
            + TUStringBuilder.concatenate(sequenceData, "<space>"));
             */
            
            if(sequenceData.length == 0
                    || (sequenceData.length == 1
                    && sequenceData[0].equals("")))
                continue;
            
            Vector tempSequence = new Vector();
            for(int j = 0; j < sequenceData.length; j++)
            {
                try
                {
                    if(sequenceData[j].trim().equals(""))
                        continue;
                    
                    //System.out.println("adding: " + sequenceData[j]);
                    tempSequence.add(sequenceData[j]);
                    //sequence[j] = Integer.valueOf(sequenceData[j]);
                }
                catch(NumberFormatException e)
                {
                    System.err.println("Read in a bad element in the input file" +
                            " at line: " + i + " check the: " + (j + 1) + "th item.");
                }
                
            }
            
            /* convert the vector back over to an array */
            String[] sequence = new String[tempSequence.size()];
            for(int k = 0; k < tempSequence.size(); k++)
                sequence[k] = (String)tempSequence.get(k);
            
            temp.add(new TUClusterableItem(sequence, i));
            /*
            System.err.println("Done adding supplemental sequence " + i + " in: "
                    + TUStringBuilder.millisecondsToHoursMinutesSeconds
                    (System.currentTimeMillis() - a));
             */
        }
        int idOfClusterToAddItemTo = -1;
        supplementalClusterableItems = new TUClusterableItem[temp.size()];
        /* now determine where the new sequences need to be placed and place them there */
        /* also don't forget to swap ids with sequences that are alread there */
        for(int i = 0; i < temp.size(); i++)
        {
            long a = System.currentTimeMillis();
            TUClusterableItem currentItem = (TUClusterableItem)temp.get(i);
            // now see where to place this item
            double shortestDistanceSoFar = Double.MAX_VALUE;
            TUCluster clusterToAddItemTo = null;
       
            long b = System.currentTimeMillis();
            for(int j = 0; j < clusters.length; j++)
            {
                TUCluster cluster = clusters[j];
                long d = System.currentTimeMillis();
                TUClusterableItem medoid = (TUClusterableItem)cluster.getMedoid
                        (TUDistanceMeasures.LONGEST_COMMON_SUBSEQUENCE);
                /*
                        System.err.println("Got the medoid in: "
                    + TUStringBuilder.millisecondsToHoursMinutesSeconds
                    (System.currentTimeMillis() - d));                 
*/
                        long e = System.currentTimeMillis();
                    double distance = TUDistanceMeasures.getDistance((Object[])medoid.getValue()
                    , (Object[])currentItem.getValue(), TUDistanceMeasures.LONGEST_COMMON_SUBSEQUENCE);
                    
                    /*
                        System.err.println("Got the distance in: "
                    + TUStringBuilder.millisecondsToHoursMinutesSeconds
                    (System.currentTimeMillis() - e));                     
                     */
                    if(distance < shortestDistanceSoFar)
                    {
                        shortestDistanceSoFar = distance;
                        idOfClusterToAddItemTo = j;
                        clusterToAddItemTo = clusters[j];
                    }
             }     
            /*
                        System.err.println("Determined which cluster to add item to in: "
                    + TUStringBuilder.millisecondsToHoursMinutesSeconds
                    (System.currentTimeMillis() - b));            
            */
            // now we can add the item to the cluster, but first we need to fix id issues.
            boolean done = false;
            b = System.currentTimeMillis();
            for(int j = 0; j < clusters.length; j++)
            {
                if(done)
                    break;
                TUCluster cluster = clusters[j];
                for(int k = 0; k < cluster.size(); k++)
                {
                    if(done)
                        break;
                    TUClusterableItem clusterItem = (TUClusterableItem)cluster.get(k);
                    if(clusterItem.getId() == currentItem.getId())
                    {
                            // swap the IDs
                            int oldId = clusterItem.getId();
                            clusterItem.setId(sequences.length + currentItem.getId());
                            //System.err.println("swapped id: " + oldId + " with id: " + clusterItem.getId());
                            done = true;                 
                    }
                }
            }        
            /*
                        System.err.println("IDs swapped in: "
                    + TUStringBuilder.millisecondsToHoursMinutesSeconds
                    (System.currentTimeMillis() - b));
            */
            // finally we add the item to the appropriate cluster
            currentItem.setCustomData(turkov.util.TUClusterUtil.SUPPLEMENTAL_ITEM,
                    turkov.util.TUClusterUtil.TRUE);
            currentItem.setOutlierScore(shortestDistanceSoFar);
            boolean resetMedoid = false;
            clusterToAddItemTo.add(currentItem, resetMedoid);              
                //System.err.println("added seq with id: " + currentItem.getId()
                   // + " to cluster with id: " + idOfClusterToAddItemTo);         
                
            supplementalClusterableItems[i] = (TUClusterableItem)temp.get(i);
            /*
                        System.err.println("Sequence " + i + " and appropriate IDs swapped in: "
                    + TUStringBuilder.millisecondsToHoursMinutesSeconds
                    (System.currentTimeMillis() - a));
             */
        }
    }    
    
    private static int[] getSequenceIdList(File sequenceIdFile)
    throws IOException
    {
        String[] sequenceIdCollections = TUStringBuilder.generateCollectionFromFile(sequenceIdFile);
        TUSet set = new TUSet();
        for(int j = 0; j < sequenceIdCollections.length; j++)
        {
            String[] sequenceIdStrings = sequenceIdCollections[j].split(" ");
            for(int i = 0; i < sequenceIdStrings.length; i++)
            {
                String itemToAdd = sequenceIdStrings[i].trim();
                Integer supposedInteger = null;
                
                try
                {
                    if(!itemToAdd.equals(""))
                    {
                        supposedInteger = Integer.parseInt(sequenceIdStrings[i].trim()) - 1;
                        if(supposedInteger.intValue() > sequences.length)
                            println("Sequence not found: " + (supposedInteger.intValue() + 1) + "!");
                        else
                            set.add(supposedInteger);
                    }
                }
                
                catch(NumberFormatException e)
                {
                    println("Invalid sequence ID specified in sequenceIdFile: "
                            + e.getMessage().split("\\:")[1].trim());
                }             
            }
        }
        
        int[] result = new int[set.size()];
        for(int i = 0; i < set.size(); i++)
            result[i] = (Integer)set.get(i);        
        
        return result;
    }
    
    private static boolean userSpecifiedAnOutlierFraction = false;
    private static TUClusterableItem[] clusterableItems;
    private static TUClusterableItem[] supplementalClusterableItems;
    private static TUCluster[] clusters;
    private static HashMap uniqueElementsToIds;
    private static HashMap idsToUniqueElements;
    private static File inputFile;
    private static File clusteringOutputFile;
    private static int numberOfPAMIterations;
    private static int numberOfClaraClusters;
    private static int numberOfCLARAIterations;
    private static double sampleFraction;
    private static double outlierFraction;
    private static String[] sequences;
    private static String[] supplementalSequences;
    private static String alterationsFileNamePrefix;
    private static boolean onlyDoClustering;    
    private static TUCSVFile predeterminedClustersCSVFile;
    private static File sequenceIdFile;
    private static File supplementalSequencesFile;
    


    private final static String DEFAULT_ALTERATIONS_FILE_NAME_PREFIX = "seq";
    private final static int DEFAULT_NUMBER_OF_PAM_ITERATIONS = 5;
    private final static int DEFAULT_NUMBER_OF_CLUSTERS = 4;
    private final static int DEFAULT_NUMBER_OF_CLARA_ITERATIONS = 5;
    private final static double DEFAULT_SAMPLE_FRACTION = 0.1;
    private final static double DEFAULT_OUTLIER_FRACTION = 0.1;
    
}
